/*=====================================================================*/
/*    serrano/prgm/project/bigloo/fthread/src/BglThread.java           */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Wed Jan 30 10:40:23 2002                          */
/*    Last change :  Thu Dec 19 09:41:41 2002 (serrano)                */
/*    Copyright   :  2002 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    Bigloo Class for threads. Nothing fancy here. Bigloo threads are */
/*    regular threads augmented with two slots:                        */
/*      - a slot used to store a back pointer to the Bigloo objects    */
/*        representing the user thread.                                */
/*      - a slot pointing to the closure implementing the body of the  */
/*        thread.                                                      */
/*=====================================================================*/
using System;
using System.Threading;
using bigloo;

namespace bigloo.fairthread
{
  /*---------------------------------------------------------------------*/
  /*    BglThread                                                        */
  /*---------------------------------------------------------------------*/
  public class BglThread
  {
    // Bigloo connection
    private Object bgl_obj;
    private Object detach_obj;
    private procedure bgl_thunk;
    private static bgldynamic current_dynamic_env;
    public bgldynamic env;

    // Public lock
    public Object _lock;

    // parent thread (i.e. the Java thread that has entered this Bgl thread)
    private Thread parent;

    // The associated thread
    private readonly Thread thread;
     
    // wait/notify synchronization
    private bool hascpup = false;
    private bool started = false;

    // public constructor
    public BglThread()
    {
      _lock = new Object();
      env = new bgldynamic( bgldynamic.current_dynamic_env );
      thread= new Thread( new ThreadStart( run ) );
    }
     
    // public constructor
    public BglThread( procedure thunk )
    {
      _lock = new Object();
      env = new bgldynamic( bgldynamic.current_dynamic_env );
      bgl_thunk = thunk;
      thread= new Thread( new ThreadStart( run ) );
    }

    // public constructor
    public BglThread( procedure thunk, symbol name )
    {
      _lock = new Object();
      env = new bgldynamic( bgldynamic.current_dynamic_env );
      bgl_thunk = thunk;
      thread= new Thread( new ThreadStart( run ) );
      thread.Name= foreign.newstring( name.pname );
    }

    // This method is invoked by Bigloo when the Bigloo object
    // representing the thread has been allocated.
    public void setup( Object o )  
    {
      bgl_obj = o;
    }

    // This method is used for compatibility with the Posix implementation
    // Don't remove it
    public void detachIdSet( Object o ) 
    {
     detach_obj = o;
    }

    // Return back the pointer to the Bigloo object
    public Object getBglObj() 
    {
      return bgl_obj;
    }

    // Return back the body of the Java thread
    public Object getThunk() 
    {
      return bgl_thunk;
    }

    // The thread entry-point
    private void run() 
    {
      try 
      {
        hascpup = true;

        lock( this ) 
        {
          started = true;
          Monitor.Wait( this );
          bgl_thunk.funcall0();
        }
      } 
      catch( Exception e ) 
      {
        Console.Error.WriteLine( e.StackTrace );
        try 
        {
          foreign.internalerror( e );
        } 
        catch( Exception ) 
        {
          Environment.Exit( 1 );
        }
      }
    }

    // await a notification
    public void await( BglThread s ) 
    {
      lock( s._lock ) 
      {
        try
        {
          Monitor.Wait( s._lock );
        } 
        catch( ThreadInterruptedException e ) 
        { 
          Console.Error.WriteLine( "interrupted wait exception: " + e ); 
        }
      }
    }
     
    // private method to get the CPU
    private void getCpu() 
    {
      hascpup = true;
      Monitor.Pulse( this );
    }
     
    // private method to loose the CPU
    private void looseCpu() 
    {
      while( hascpup == false ) 
      {
        try
        {
          Monitor.Wait( this );
        } 
        catch( ThreadInterruptedException e ) 
        { 
          Console.Error.WriteLine( "interrupted wait exception: " + e ); 
        }
      }
    }
     
    //public wrapper for looseCpu
    public void waitForToken() {
      looseCpu();
    }

    // Switch to another Bglthread and don't wait for the cpu
    public void switchExit( BglThread next ) 
    {
      hascpup = false;
        
      while( next.started == false )
        Thread.Sleep( 0 );

      lock( this ) 
      {
        lock( next ) 
        {
          bgldynamic.current_dynamic_env = next.env;
          next.getCpu();
        }
      }
    }
     
    // Switch to another Bglthread
    public void switchTo( BglThread next ) 
    {
      switchExit( next );
      looseCpu();
    }
     
    /*    // Switch to another Bglthread                                   */
    /*    public void switchTo( BglThread next ) {                         */
    /*       hascpup = false;                                              */
    /*                                                                     */
    /*       while( next.started == false ) Thread.yield();                */
    /*                                                                     */
    /*       synchronized( this ) {                                        */
    /* 	 synchronized( next ) {                                        */
    /* 	    bgldynamic.current_dynamic_env = next.env;                 */
    /* 	    next.getCpu();                                             */
    /* 	 }                                                             */
    /* 	 looseCpu();                                                   */
    /*       }                                                             */
    /*    }                                                                */
    /*                                                                     */
    // Enter a Bglthread from a regular thread
    static public void switchIn( BglThread next ) 
    {
      next.parent = Thread.CurrentThread;
        
      while( !next.started )
        Thread.Sleep( 0 );

      lock( next.parent ) 
      {
        lock( next ) 
        {
          current_dynamic_env = bgldynamic.current_dynamic_env;
          bgldynamic.current_dynamic_env = next.env;
          next.getCpu();
        }
        try
        {
          Monitor.Wait( next.parent );
        } 
        catch( ThreadInterruptedException e ) 
        { 
          Console.Error.WriteLine( "interrupted wait exception: " + e ); 
        }
      }
    }
     
    // Leave a Bglthread for a regular thread
    public void switchOut() 
    {
      lock( parent )
      {
        bgldynamic.current_dynamic_env = current_dynamic_env;
        Monitor.Pulse( parent );
      }
      lock( this )
      {
        try
        {
          Monitor.Wait( this );
        } 
        catch( ThreadInterruptedException e ) 
        {
          Console.Error.WriteLine( "interrupted wait exception: " + e );
        }
      }
    }

    // Starts the thread
    public void start()
    {
      thread.Start();
    }
  }
}

